#!/usr/bin/env python
from __future__ import print_function
import os
import sys
import struct
import subprocess as sp


"""gen-dbpl.py: create DeaDBeeF playlists from a list of media files"""
__version__   = "1.2"
__author__    = "ed <gendbpl@ocv.me>"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2018


def p1buf(s):
	""" pascal string, max 256 chars """
	return struct.pack('B', len(s)) + s


def p2buf(s):
	""" pascal string, max 65536 chars """
	return struct.pack('<H', len(s)) + s


def p1str(s):
	""" pascal string, max 256 chars """
	return p1buf(s.encode('utf-8'))


def p2str(s):
	""" pascal string, max 65536 chars """
	return p2buf(s.encode('utf-8'))


def ffmpeg(fn):
	with open(os.devnull, 'r+b') as null:
		ff = sp.Popen([b'ffprobe', b'-hide_banner', b'--', fn],
			stdin=null, stdout=null, stderr=sp.PIPE)
		
		return ff.communicate()[1].decode('utf-8', 'replace')


def get_metadata(media_path):
	ff = ffmpeg(media_path)
	tags_ind = -1
	tags = {}
	maxdur = -1
	for ln in ff.split('\n'):
		ln = ln.rstrip('\r')
		sln = ln.lstrip(' ')
		
		if sln.startswith('Duration:'):
			dur = sln.split(' ')[1][:-1].split(':')
			try:
				dur = 60*(60*int(dur[0])+int(dur[1]))+float(dur[2])
				if maxdur < dur:
					maxdur = dur
			except:
				pass
			
			tags_ind = -1
			continue
		
		ind = len(ln) - len(sln)
		if ind < tags_ind:
			tags_ind = -1
		
		if sln == 'Metadata:':
			tags_ind = ind + 2
			continue
		
		if ind == tags_ind:
			k, v = [x.strip() for x in ln.split(':', 1)]
			tags[k] = v
	
	return [maxdur, tags]


def bin_stdin():
	if sys.version_info[0] > 2:
		stdin = sys.stdin.buffer
	else:
		if sys.platform == 'win32':
			import os, msvcrt
			msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
		stdin = sys.stdin

	sibuf = b''
	while True:
		d = stdin.read(4096)
		if not d:
			break
		sibuf += d
		while True:
			ofs = sibuf.find(b'\n')
			if ofs < 0:
				break
			rv = sibuf[:ofs]
			sibuf = sibuf[ofs+1:]
			yield rv


def make_dbpl(dbpl):
	abspath = b''
	if len(sys.argv) > 2:
		abspath = sys.argv[2].encode('utf-8')
		if not abspath.endswith(b'/'):
			abspath += b'/'

	# magic(DBPL) verMajor(1) verMinor(2) numFiles(uint)
	dbpl.write(b'\x44\x42\x50\x4c\x01\x02\x00\x00\x00\x00')

	n_items = 0
	for f in [x.rstrip(b'\r\n') for x in bin_stdin()]:
		if f.startswith(b'./'):
			f = f[2:]

		md = get_metadata(f)
		dur = md[0]
		if dur < 10:
			dur = 300
		
		print(u'{} \033[33m{:7.2f} \033[36m{}\033[0m'.format(
			n_items, dur, f.decode('utf-8', 'replace')))
		
		buf = p2buf(abspath + f)
		buf += p1str('ffmpeg')
		buf += struct.pack('<HLLf',
			0,   # track number
			0,   # start sample
			0,   # end sample
			dur) # duration
		
		ext = f.split(b'.')[-1]
		if len(ext) < 2 \
		or len(ext) > 6:
			ext = b'mp3'
		
		buf += p1buf(ext)
		
		buf += struct.pack('<LfLfLH',
			0,   # album gain
			0,   # album peak
			0,   # track gain
			0,   # track peak
			0,   # flags
			4)   # num tags
		
		buf += p2str(':URI')
		buf += p2buf(abspath + f)  # does not appear to be escaped
		buf += p2str('DECODER')
		buf += p2str('ffmpeg')
		buf += p2str('FILETYPE')
		buf += p2buf(ext)
		buf += p2str('FILE_SIZE')
		buf += p2str(str(os.path.getsize(f)))
		
		dbpl.write(buf)
		n_items += 1
	
	dbpl.write(b'\x00\x00')  # TODO: ?
	dbpl.seek(6)
	dbpl.write(struct.pack('<L', n_items))


def test_ffmpeg():
	md = get_metadata(sys.argv[1])
	print('\033[36mDuration: \033[0m{}\n\033[36mTags:'.format(md[0]))
	for k, v in sorted(md[1].items()):
		print('  \033[33m{:<13} \033[0m{}'.format(k, v))


def main():
	#v = b'\x3f\x80\x00\x00'
	#print('{:f}'.format(struct.unpack('f',v)[0]))
	
	#test_ffmpeg()
	#return

	if len(sys.argv) < 2:
		print()
		print('  expecting list of audio file paths on stdin')
		print()
		print('  need argument 1:  path/to/output.dbpl')
		print()
		print('  optional arg. 2:  absolute path prefix;')
		print('     the relative path of each playlist entry')
		print('     will be appended to this argument,')
		print('     forming the absolute path')
		print()
		print('  example usage:')
		print('    find -iname \*.ogg | sort | gen-dbpl.py the.dbpl "$(pwd)')
		print()
		sys.exit(1)

	with open(sys.argv[1], 'wb') as f:
		make_dbpl(f)


if __name__ == '__main__':
	main()
else:
	print('excuse me?', file=sys.stderr)


"""
s name
s decoder_id
2 track num
4 startsample
4 endsample
4 duration
s filetype
4 albumgain
4 albumpeak
4 trackgain
4 trackpeak
4 flags
x metadata
"""
